#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
import datetime
import os
import subprocess
import shutil
import sys

HERE = os.path.dirname(__file__)
sys.path.append(os.path.join(HERE, '../../lib'))

from charmhelpers import (
    core,
    fetch,
)

CHROOT_USER_HOME = '/var/lib/jenkins'
CHROOT_ROOT_PATH = '/var/cache/pbuilder'
REQUIRED_PACKAGES = [
    'apt-transport-https',
    'build-essential',
    'bzr',
    'bzr-builddeb',
    'ccache',
    'cdbs',
    'cli-common-dev',
    'click',
    'click-dev',
    'curl',
    'debootstrap',
    'devscripts',
    'dh-autoreconf',
    'dh-golang',
    'dh-migrations',
    'dh-systemd',
    'dh-translations',
    'dput',
    'eatmydata',
    'gnome-pkg-tools',
    'gobject-introspection',
    'gtk-doc-tools',
    'intltool',
    'java-common',
    'libgconf2-dev',
    'libglib2.0-dev',
    'libgtk-3-dev',
    'libgtk2.0-dev',
    'mozilla-devscripts',
    'pbuilder',
    'pbuilder-scripts',
    'pkg-kde-tools',
    'python-distutils-extra',
    'python-setuptools',
    'python3-distutils-extra',
    'python3-setuptools',
    'qemu-user-static',
    'quilt',
    'sphinx-common',
    'ubuntu-dev-tools',
    'zip',
]
CHROOT_EXTRA_PACKAGES = [
    'apt-transport-https',
    'ca-certificates',
    'devscripts',
    'eatmydata',
    'gnupg',
    'patchutils',
    'pbuilder',
    'vim-tiny',
]


def juju_info(msg, level='INFO'):
    core.hookenv.log(msg, level)


def chroot_exists(path, dist, arch):
    chroot_path = os.path.join(path, '{}-{}.tgz'.format(dist, arch))
    return os.path.isfile(chroot_path)


def get_chroot_from_url(path, dist, arch, url):
    if not url:
        return False
    chroot_path = os.path.join(path, '{}-{}'.format(dist, arch))
    wget_path = os.path.join(path, '{}-{}.tgz'.format(dist, arch))
    url_path = '{}/{}-{}.tgz'.format(url, dist, arch)
    try:
        subprocess.check_call(['wget', '-O', wget_path, url_path])
        subprocess.check_call(['tar', '-xzf', wget_path, '-C', path])
    except subprocess.CalledProcessError:
        juju_info('Could not download or extract {}'.format(url_path),
                  'WARNING')
        # Clean up
        try:
            subprocess.check_call(['rm', '-rf', chroot_path])
            subprocess.check_call(['rm', '-rf', wget_path])
        except subprocess.CalledProcessError:
            raise EnvironmentError(
                'Could not cleanup chroot {}'.format(chroot_path))
        return False
    return True


def update_chroot(path, dist, arch):
    chroot_path = os.path.join(path, '{}-{}'.format(dist, arch))
    chroot_env = copy.copy(os.environ)
    chroot_env['DIST'] = dist
    chroot_env['ARCH'] = arch
    chroot_env['HOME'] = CHROOT_USER_HOME
    try:
        start = datetime.datetime.now()
        subprocess.check_call(['pbuilder', '--update'],
                              env=chroot_env)
        end = datetime.datetime.now()
        juju_info('chroot update: {}'.format(end - start))
    except subprocess.CalledProcessError:
        raise EnvironmentError(
            'Could not update chroot {}'.format(chroot_path))


def create_chroot(path, dist, arch):
    '''Creates a pbuilder chroot given the dist and arch.'''
    chroot_env = copy.copy(os.environ)
    chroot_env['DIST'] = dist
    chroot_env['ARCH'] = arch
    chroot_env['HOME'] = CHROOT_USER_HOME
    start = datetime.datetime.now()
    juju_info('Attempting: pbuilder --create for {}-{}'.format(dist, arch))
    try:
        subprocess.check_call(['pbuilder', '--create',
                               '--distribution', dist,
                               '--architecture', arch,
                               '--debootstrapopts',
                               '--variant=buildd',
                               '--extrapackages',
                               ' '.join(CHROOT_EXTRA_PACKAGES),
                              ],
                              env=chroot_env)
        #subprocess.check_call(['pbuilder', '--save',
        #                       '--execute', '--', '/usr/bin/apt-get',
        #                       'install', 'eatmydata', 'bzr-builddeb',
        #                       'software-properties-common', '-yq',
        #                       '--force-yes'],
        #                      env=chroot_env)
    except subprocess.CalledProcessError as e:
        raise EnvironmentError(
            'Could not create chroot for {}-{} return: {}'.format(
                dist, arch, e.returncode))
    end = datetime.datetime.now()
    juju_info('chroot create: {}'.format(end - start))


def setup_chroots(config):
    distros = config.get('distros', None)
    architectures = config.get('architectures', None)
    base_url = config.get('chroot_base_url', '')

    try:
        dist_list = distros.split()
    except:
        juju_info('''Can't determine distros from [{}]'''.format(distros),
                  'ERROR')
        return 1

    try:
        architecture_list = architectures.split()
    except:
        juju_info('''Can't determine architectures from [{}]'''.format(
            architectures), 'ERROR')
        return 1

    for dist in dist_list:
        for arch in architecture_list:
            try:
                if chroot_exists(CHROOT_ROOT_PATH, dist, arch):
                    update_chroot(CHROOT_ROOT_PATH, dist, arch)
                elif get_chroot_from_url(CHROOT_ROOT_PATH, dist, arch,
                                         base_url):
                    update_chroot(CHROOT_ROOT_PATH, dist, arch)
                else:
                    create_chroot(CHROOT_ROOT_PATH, dist, arch)
            except EnvironmentError as e:
                juju_info('{}'.format(e), 'ERROR')
                return 1
    return 0


def setup_dependencies(config):
    fetch.apt_update()
    juju_info('installing apt packages...')
    fetch.apt_install(REQUIRED_PACKAGES, fatal=True)


def setup_pbuilderrc(config):
    shutil.copy(os.path.join(HERE, '..', '..', 'files', 'pbuilder',
                             '.pbuilderrc'),
                CHROOT_USER_HOME)


def setup(config):
    distros = config.get('distros', None)
    architectures = config.get('architectures', None)
    if not distros or not architectures:
        juju_info('Skipping chroot setup distros: [{}], '
                  'archs: [{}]'.format(distros, architectures))
        return 0
    setup_dependencies(config)
    setup_pbuilderrc(config)
    setup_chroots(config)


def main():
    juju_info('Running chroot setup: {}'.format(os.environ.get('JUJU_HOOK_NAME')))

    config = core.hookenv.config()
    juju_info('Config: {}'.format(config))
    if sys.argv[1] in ['install', 'config-changed', 'upgrade-charm']:
        try:
            return setup(config)
        except subprocess.CalledProcessError as e:
            juju_info('Error running: {}: {}'.format(e.cmd, e.output))
            return e.returncode


if __name__ == '__main__':
    exit(main())
